"use strict";

this.EXPORTED_SYMBOLS = [
  "UnifiedUrlbar"
];

const STYLE_URL = "chrome://unified-urlbar/content/style.css";
const SEARCH_BAR_WIDGET_ID = "search-container";
const XHTML_NS = "http://www.w3.org/1999/xhtml";

const { classes: Cc, interfaces: Ci, utils: Cu } = Components;
Cu.import("resource://gre/modules/XPCOMUtils.jsm");
Cu.import("resource://gre/modules/Services.jsm");
Cu.import("resource://gre/modules/Preferences.jsm");
Cu.import("resource:///modules/CustomizableUI.jsm");
Cu.import("resource:///modules/BrowserUITelemetry.jsm");

var gOptions = {
    removeSearchbar: true,
    resultsLimit: false,
    matchBuckets: "",
    matchBucketsSearch: "",
};

var gBrowsers = null;
var gPrefSvc = null;

function mirrorPref(name, defaultValue, newValue) {
  let mirrorName = "browser.urlbar.experiment.unified-urlbar.mirror." + name;
  if (!gPrefSvc.isSet(mirrorName)) {
    gPrefSvc.set(mirrorName, gPrefSvc.get(name, defaultValue));
    if (!gPrefSvc.isSet(name)) {
      gPrefSvc.set(mirrorName + ".notset", true);
    }
  }
  gPrefSvc.set(name, newValue);
}

function restoreMirroredPref(name) {
  // This is backwards compatible with ".notset" not being initialized.
  let mirrorName = "browser.urlbar.experiment.unified-urlbar.mirror." + name;
  if (gPrefSvc.isSet(mirrorName)) {
    if (gPrefSvc.isSet(mirrorName + ".notset")) {
      gPrefSvc.reset(name);
    } else {
      gPrefSvc.set(name, gPrefSvc.get(mirrorName));
    }
    gPrefSvc.reset(mirrorName);
    gPrefSvc.reset(mirrorName + ".notset");
  }
}

this.UnifiedUrlbar = Object.freeze({
  isUserEligible() {
    // This experiment should be installable on Firefox 55 and 56.
    let vc = Cc["@mozilla.org/xpcom/version-comparator;1"]
               .getService(Ci.nsIVersionComparator);
    let appInfo = Cc["@mozilla.org/xre/app-info;1"]
                    .getService(Ci.nsIXULAppInfo);
    if (vc.compare(appInfo.version, "55") < 0) {
      return false;
    }

    // Exclude users who already removed the search bar from the UI.
    if (CustomizableUI.getPlacementOfWidget("search-container") === null) {
      return false;
    }

    return true;
  },

  init(options = gOptions, prefSvc) {
    if (gBrowsers) {
      Cu.reportError("UnifiedUrlbar.init() was invoked multiple times?");
      return;
    }

    gOptions = options;
    gPrefSvc = prefSvc;

    // Restoring prefs allows testing by changing the branch and restarting.
    if (gOptions.resultsLimit) {
      mirrorPref("browser.urlbar.maxRichResults", 10, gOptions.resultsLimit);
    } else {
      restoreMirroredPref("browser.urlbar.maxRichResults");
    }
    if (gOptions.matchBuckets) {
      mirrorPref("browser.urlbar.matchBuckets", "", gOptions.matchBuckets);
    } else {
      restoreMirroredPref("browser.urlbar.matchBuckets");
    }
    if (gOptions.matchBucketsSearch) {
      mirrorPref("browser.urlbar.matchBucketsSearch", "", gOptions.matchBucketsSearch);
    } else {
      restoreMirroredPref("browser.urlbar.matchBucketsSearch");
    }

    gBrowsers = new Set();
    getBrowserWindows();
    Services.ww.registerNotification(this);
  },

  diagnostics() {
    let diagnostics = this.searchBarDiagnostics();
    let getPreference = function (name, fn, prefName) {
      try {
        diagnostics[name] = Services.prefs[fn](prefName);
      } catch (ex) {}
    };
    getPreference("oneoff", "getBoolPref", "browser.urlbar.oneOffSearches");
    getPreference("suggestions", "getBoolPref", "browser.urlbar.suggest.searches");
    getPreference("maxRichResults", "getIntPref", "browser.urlbar.maxRichResults");
    getPreference("matchBuckets", "getCharPref", "browser.urlbar.matchBuckets");
    getPreference("matchBucketsSearch", "getCharPref", "browser.urlbar.matchBucketsSearch");
    return diagnostics;
  },

  searchBarDiagnostics() {
    let win = Services.wm.getMostRecentWindow("navigator:browser");
    if (!win) {
      return { allWindowsClosed: true };
    }

    let elt = win.document.getElementById("search-container");
    if (!elt) {
      return { searchBarRemovedManually: true };
    }

    let width = elt.clientWidth;
    if (!width) {
      return { searchBarRemovedByExperiment: true };
    }

    return { searchBarWidth: width };
  },

  destroy(prefSvc) {
    gPrefSvc = prefSvc;

    restoreMirroredPref("browser.urlbar.maxRichResults");
    restoreMirroredPref("browser.urlbar.matchBuckets");
    restoreMirroredPref("browser.urlbar.matchBucketsSearch");

    // Also restore changes from previous versions of the experiment.
    restoreMirroredPref("browser.urlbar.suggest.searches");
    restoreMirroredPref("browser.urlbar.oneOffSearches");

    if (!gBrowsers) {
      // This was called on startup before "init", because the study ended
      return;
    }

    Services.ww.unregisterNotification(this);
    for (let browser of gBrowsers) {
      browser.destroy();
    }
    gBrowsers.clear();
    gBrowsers = null;
  },

  observe(subject, topic, data) {
    if (topic == "autocomplete-did-enter-text") {
      trackAutocompleteEnter(subject.QueryInterface(Ci.nsIAutoCompleteInput));
      return;
    }

    let win = subject.QueryInterface(Ci.nsIDOMWindow);
    if (!win) {
      return;
    }

    if (topic == "domwindowopened") {
      whenWindowLoaded(win, () => {
        if (isValidBrowserWindow(win)) {
          gBrowsers.add(new Browser(win));
        }
      });
    } else if (topic == "domwindowclosed") {
      for (let browser of gBrowsers) {
        if (browser.window == win) {
          browser.destroy();
          gBrowsers.delete(browser);
          break;
        }
      }
    }
  },

  reportTelemetryValue(key, optionalData={}) {
    reportTelemetryValue(key, optionalData);
  }
});

function Browser(win, branch) {
  this.window = win;

  // Per window toggles.
  if (("removeSearchbar" in gOptions) && gOptions.removeSearchbar) {
    this._removeSearchBar();
    win.gNavToolbox.addEventListener("customizationstarting", this);
    win.gNavToolbox.addEventListener("customizationending", this);
  }

  win.gBrowser.addProgressListener(this);
}

Browser.prototype = {
  get document() {
    return this.window.document;
  },

  destroy() {
    this.window.gBrowser.removeProgressListener(this);

    this.window.gNavToolbox.removeEventListener("customizationstarting", this);
    this.window.gNavToolbox.removeEventListener("customizationending", this);
    this._restoreSearchBar();
  },

  handleEvent(event) {
    switch (event.type) {
      case "customizationstarting":
        this._restoreSearchBar();
        break;
      case "customizationending":
        this._removeSearchBar();
        break;
    }
  },

  _removeSearchBar() {
    let elt = this.document.getElementById("search-container");
    if (elt) {
      elt.style.display = "none";
    }
  },

  _restoreSearchBar() {
    let elt = this.document.getElementById("search-container");
    if (elt) {
      elt.style.removeProperty("display");
    }
  },

  onLocationChange(webProgress, request, uri, flags) {
    try {
      if (webProgress.isTopLevel && uri.host) {
        let host = uri.host.replace(/^www./, "").replace(/^search./, "");
        if (gEngines.has(host)) {
          let rv = Services.search.parseSubmissionURL(uri.spec);
          // HACK: try to not count result pages we generated and subpages.
          // This is tricky and working until the engines keep same param names.
          if (rv.engine &&
              !["hspart=mozilla", // Yahoo tracking
                "&b=",            // Yahoo paging
                "client=firefox", // Google tracking
                "&start=",        // Google paging
                "pc=MOZI",        // Bing tracking
                "&first=",        // Bing paging
               ].some(str => uri.path.includes(str))) {
            reportTelemetryValue("userVisitedEngineResult",
                                 { engine: rv.engine });
          }
        }
      }
    } catch (ex) {}
  },
  onProgressChange() {},
  onSecurityChange() {},
  onStateChange(webProgress, request, flags, status) {
    try {
      if (webProgress.isTopLevel &&
          flags & Ci.nsIWebProgressListener.STATE_START &&
          flags & Ci.nsIWebProgressListener.STATE_IS_NETWORK &&
          (request && (request instanceof Ci.nsIChannel || "URI" in request)) &&
          request.URI.path == "/") {
        let host = request.URI.host.replace(/^www./, "").replace(/^search./, "");
        if (gEngines.has(host)) {
          reportTelemetryValue("userVisitedEngineHost",
                               { engine: gEngines.get(host) });
        }
      }
    } catch (ex) {}
  },
  onStatusChange() {},

  QueryInterface: XPCOMUtils.generateQI([ Ci.nsIWebProgressListener ])
};

XPCOMUtils.defineLazyGetter(this, "gEngines", () => {
  let engines = new Map();
  for (let engineName of [ "Google", "Yahoo", "Bing"]) {
    let engine = Services.search.getEngineByName(engineName);
    if (engine) {
      try {
        let engineHost = Services.io.newURI(engine.searchForm, null, null).host;
        engines.set(engineHost.replace(/^www./, "").replace(/^search./, ""),
                    engine);
      } catch (ex) {}
    }
  }
  return engines;
});

function isValidBrowserWindow(win) {
  return !win.closed && win.toolbar.visible &&
          win.document.documentElement.getAttribute("windowtype") == "navigator:browser";
}

function getBrowserWindows() {
  let wins = Services.ww.getWindowEnumerator();
  while (wins.hasMoreElements()) {
    let win = wins.getNext().QueryInterface(Ci.nsIDOMWindow);
    whenWindowLoaded(win, () => {
      if (isValidBrowserWindow(win)) {
        gBrowsers.add(new Browser(win));
      }
    });
  }
}

function whenWindowLoaded(win, callback) {
  if (win.document.readyState == "complete") {
    callback();
    return;
  }
  win.addEventListener("load", function onLoad(event) {
    if (event.target == win.document) {
      win.removeEventListener("load", onLoad, true);
      win.setTimeout(callback, 0);
    }
  }, true);
}

function trackAutocompleteEnter(input) {
  if (!input || input.id != "urlbar" || input.inPrivateContext) {
    return;
  }

  let controller = input.popup.view.QueryInterface(Ci.nsIAutoCompleteController);

  let idx = input.popup.selectedIndex;
  if (idx == -1) {
    return;
  }
  let value = controller.getValueAt(idx);
  let action = input._parseActionUrl(value);
  if (action.type != "searchengine")
    return;

  if ("searchSuggestion" in action.params) {
    reportTelemetryValue("searchBySuggestion");
  } else if ("alias" in action.params) {
    let engine = Services.search.getEngineByAlias(action.params.alias);
    if (engine)
      reportTelemetryValue("searchByAlias", { engine });
  } else {
    reportTelemetryValue("searchByDefaultEngine");
  }
}

/**
 * Registers the presence of an event.
 *
 * @param eventName The data is logged with this name.
 */
function reportTelemetryValue(key, optionalData={}) {
  // Since we only care about search volume, ignore the rare cases where a
  // search opens in a new tab (currently onlye CTRL + Go button) and always
  // set where to "current".
  function recordOneOff(engine, source="unknown", type="unknown") {
      let engineId = engine ? engine.identifier ? engine.identifier
                                                : "other-" + engine.name
                            : "other";
      BrowserUITelemetry.countOneoffSearchEvent(`${engineId}.${source}`, type,
                                                "current");
  }

  // Increment "search" / "urlbar" countable.
  // Since we only care about search volume, don't report details, like the
  // suggestion index clicked, to countSearchEvent.
  function recordSearch(engine=Services.search.currentEngine, source) {
    let isInContentSearch = source.includes("content");
    if (isInContentSearch) {
      // Add engine name to the probe name.
      source += "-" + engine.name.toLowerCase();
    }

    BrowserUITelemetry.countSearchEvent(source, null);

    if (!isInContentSearch) {
      let engineId = engine ? engine.identifier ? engine.identifier
                                                : "other-" + engine.name
                            : "other";
      let count = Services.telemetry.getKeyedHistogramById("SEARCH_COUNTS");
      count.add(`${engineId}.${source}`);
    }
  }

  switch(key) {
    case "userVisitedEngineHost":
      recordSearch(optionalData.engine, "content");
      break;
    case "userVisitedEngineResult":
      recordSearch(optionalData.engine, "content-result");
      break;
    case "searchSettingsClicked":
      BrowserUITelemetry.countSearchSettingsEvent("urlbar");
      break;
    case "searchByReturnKeyOnOneOffButton":
      recordOneOff(optionalData.engine, "urlbar-oneoff", "key");
      recordSearch(optionalData.engine, "urlbar");
      break;
    case "searchByClickOnOneOffButton":
      recordOneOff(optionalData.engine, "urlbar-oneoff", "mouse");
      recordSearch(optionalData.engine, "urlbar");
      break;
    case "searchBySuggestions":
      recordSearch(optionalData.engine, "urlbar-suggestion");
      break;
    case "searchByDefaultEngine":
      // If it's a direct Enter actions it already increments "search" / "urlbar"
      // countable, so we should only increment oneoff countable for the current
      // engine (with unknown source, like the searchbar).
      recordOneOff(Services.search.currentEngine, "urlbar", "unknown");
      break;
    case "searchByAlias":
      // Due to 1313080 aliases are not accounted by default, so do that here.
      recordOneOff(optionalData.engine, "urlbar", "unknown");
      recordSearch(optionalData.engine, "urlbar");
      break;
    default:
      Cu.reportError("reportTelemetryValue() got an unknown value");
  }
}
